# vim: set et ts=4 sw=4 fileencoding=utf-8:
'''
yalp.parsers.url
================

Extract components of a url.

The parser supports the following configuration items:

*field*
    The field containing the url string to parse. If the field is not
    found in the event, the event will be skipped. Defaults to
    ``request``.

*out_field*
    The field to set the url components to. Defaults to ``url``.

*type*
    A type filter. Events not of this type will be skipped.


Example configuration.

.. code-block:: yaml

    parsers:
      - url:
          field: 'request'

With an input event like the following:

.. code-block:: python

    {
        'hostname': 'server_hostname',
        'time_stamp': '2015-01-01T01:00:00',
        'request': '/index.html?param1=val1&param2=val2',
    }

After the parser runs, the event will become:

.. code-block:: python

    {
        'hostname': 'server_hostname',
        'time_stamp': '2015-01-01T01:00:00',
        'request': '/index.html?param1=val1&param2=val2',
        'url': {
            'fragment': '',
            'hostname': None,
            'netloc': '',
            'params': '',
            'password': None,
            'path': '/index.html',
            'port': None,
            'query': {
                'param1': ['val1'],
                'param2': ['val2'],
            },
            'scheme': '',
            'username': None
        },
    }

'''
try:
    from urlparse import urlparse, parse_qs
except ImportError:
    # pylint: disable=no-name-in-module,import-error
    from urllib.parse import urlparse, parse_qs
    # pylint: enable=no-name-in-module,import-error

from yalp.parsers import ExtractFieldParser


class UrlParser(ExtractFieldParser):
    '''
    Split urls into components.
    '''
    def __init__(self,
                 field='request',
                 out_field='url',
                 *args,
                 **kwargs):
        super(UrlParser, self).__init__(field, *args, **kwargs)
        self.out_field = out_field

    def parse(self, event):
        url_str = self.data
        url_parts = urlparse(url_str)
        query_params = parse_qs(url_parts.query)
        event[self.out_field] = {
            'scheme': url_parts.scheme,
            'netloc': url_parts.netloc,
            'path': url_parts.path,
            'params': url_parts.params,
            'query': query_params,
            'fragment': url_parts.fragment,
            'username': url_parts.username,
            'password': url_parts.password,
            'hostname': url_parts.hostname,
            'port': url_parts.port,
        }
        return event


Parser = UrlParser



# vim: set et ts=4 sw=4 fileencoding=utf-8:
'''
yalp.inputs
===========
'''
import logging
from datetime import datetime

from ..config import settings
from ..utils import get_hostname
from ..pipeline import ThreadPipline

DEFAULT_DATE_FMT = '%Y-%m-%dT%H:%M:%S'


class InputerMixin(object):
    '''
    Mixin for an Inputer

    Makes it easy to enqueue events.
    '''

    def enqueue_event(self, event):
        '''
        Take an event and send it to the proper queue.

        Send the event to the parsers queue, unless there are no parsers
        in the config. Then send the event directly to the outputs.
        '''
        if self.type_ and 'type' not in event:
            event['type'] = self.type_
        event['hostname'] = self.hostname
        if 'time_stamp' not in event:
            event['time_stamp'] = datetime.now().strftime(DEFAULT_DATE_FMT)
        from yalp.pipeline import tasks
        if settings.parsers:
            tasks.process_message.apply_async(
                args=[event],
                queue=settings.parser_queue,
                serializer=settings.celery_serializer,
            )
        else:
            tasks.process_output(event)


class BaseInputer(InputerMixin, ThreadPipline):
    ''' Base Inputer '''

    def __init__(self, *args, **kwargs):
        super(BaseInputer, self).__init__(*args, **kwargs)
        self.logger = logging.getLogger(__name__)
        self.hostname = get_hostname()



# vim: set et ts=4 sw=4 fileencoding=utf-8:
'''
yalp.exceptions
===============
'''


class ImproperlyConfigured(Exception):
    '''
    Raised when configuration is not correct.
    '''


class ShutdownException(Exception):
    '''
    Raised when sigterm is sent to shutdown.
    '''


class OutputException(Exception):
    '''
    Raised when something is invalid when attempting to output an event.
    '''



# vim: set et ts=4 sw=4 fileencoding=utf-8:
'''
yalp.pipeline
=============
'''
import threading
import logging


class BasePipline(object):
    '''
    Base yalp class.
    '''
    def __init__(self, type_=None, **kwargs):  # pylint: disable=W0613
        super(BasePipline, self).__init__()
        self.type_ = type_


class CeleryPipeline(BasePipline):
    '''
    Pipeline class that iteracts with celery.
    '''
    def __init__(self, **kwargs):
        super(CeleryPipeline, self).__init__(**kwargs)
        self.logger = logging.getLogger(__name__)

    def run(self, event):
        '''
        Execute this pipeline component with the event.
        '''
        if event is None or self.type_ != event.get('type', None):
            self.logger.info('%s skipping event %s: not same type',
                             self.__class__.__name__,
                             event)
            return event
        else:
            self.logger.info('processing event %s', event)
            return self.process_event(event)

    def process_event(self, event):
        '''
        Process the event.
        '''
        raise NotImplementedError


class ThreadPipline(BasePipline, threading.Thread):
    '''
    Pileline class that is used by threads.
    '''
    def __init__(self, *args, **kwargs):
        super(ThreadPipline, self).__init__(*args, **kwargs)
        self._stopper = threading.Event()

    def stop(self):
        ''' Inform the thread to stop '''
        self._stopper.set()

    @property
    def stopped(self):
        ''' True if thread has been told to stop '''
        return self._stopper.is_set()



# vim: set et ts=4 sw=4 fileencoding=utf-8:
'''
yalp.parsers
============
'''
import logging

from ..utils import nested_get
from ..pipeline import CeleryPipeline


class BaseParser(CeleryPipeline):
    '''
    Base parser.
    '''
    def __init__(self, *args, **kwargs):
        super(BaseParser, self).__init__(*args, **kwargs)
        self.logger = logging.getLogger(__name__)

    def process_event(self, event):
        return self.parse(event)

    def parse(self, event):
        '''
        Parse the message and return event with parsed result.
        '''
        raise NotImplementedError


class ExtractFieldParser(BaseParser):
    '''
    Extracts a field from the event for parsing.

    If the field is not in the event, the event is not parsed.
    '''
    def __init__(self, field='message', *args, **kwargs):
        super(ExtractFieldParser, self).__init__(*args, **kwargs)
        self.field = field

    def process_event(self, event):
        try:
            self.data = nested_get(event, self.field)
            return super(ExtractFieldParser, self).process_event(event)
        except KeyError:
            return event



# vim: set et ts=4 sw=4 fileencoding=utf-8:
'''
yalp.utils
==========
'''
from .pipeline import BasePipline
from .exceptions import ImproperlyConfigured
from .config import settings


YALP_PLUGIN_MAPPINGS = {
    'input': {
        'classname': 'Inputer',
        'packages_list': 'input_packages',
    },
    'parser': {
        'classname': 'Parser',
        'packages_list': 'parser_packages',
    },
    'output': {
        'classname': 'Outputer',
        'packages_list': 'output_packages',
    },
}


def get_yalp_class(plugin, config, plugin_type, instance_type=BasePipline):
    '''
    Get a yalp input/parser/output class.
    '''
    class_name = YALP_PLUGIN_MAPPINGS[plugin_type]['classname']
    package_list = getattr(
        settings, YALP_PLUGIN_MAPPINGS[plugin_type]['packages_list'], [])

    for package in package_list:
        try:
            module_name = '.'.join([package, plugin])
            module = __import__(module_name, fromlist=[class_name])
            class_ = getattr(module, class_name)
            if 'type' in config:
                config['type_'] = config['type']
            instance = class_(**config)
            if not isinstance(instance, instance_type):
                raise ImportError
            return instance
        except (ImportError, NameError):
            pass
    raise ImproperlyConfigured('Invalid parser module/class.')


def get_hostname():
    '''
    Get system's hostname for worker process name.
    '''
    import socket
    hostname = socket.gethostname()
    del socket
    return hostname


def nested_get(data, lookup):
    '''
    Preform a nested dict and list lookup.
    '''
    for key in lookup.split(':'):
        if isinstance(data, list):
            try:
                idx = int(key)
                data = data[idx]
            except (ValueError, IndexError):
                raise KeyError
        else:
            try:
                data = data[key]
            except (KeyError, TypeError):
                raise KeyError
    return data


def nested_put(data, lookup, value):
    '''
    Update a field in a nested dict.
    '''
    try:
        initial_keys, last_key = lookup.rsplit(':', 1)
        data = nested_get(data, initial_keys)
    except ValueError:
        last_key = lookup
    data[last_key] = value



# vim: set et ts=4 sw=4 fileencoding=utf-8:
# pylint: disable=line-too-long
'''
yalp.parsers.user_agent
=======================

Extract browser, OS, device and other information from a user agent
string.

The parser supports the following configuration items:

*field*
    The field containing the user agent string to parse. If the field is
    not found in the event, the event will be skipped. Defaults to
    ``agent``.

*out_field*
    Set this to a field to store the user agent information under. If not
    set, the new field are added at the top level of the event.

*type*
    A type filter. Events not of this type will be skipped.


Example configuration.

.. code-block:: yaml

    parsers:
      - user_agent:
          field: 'agent'
          out_field: 'user_agent'

With an input event like the following:


.. code-block:: python

    {
        'hostname': 'server_hostname',
        'time_stamp': '2015-01-01T01:00:00',
        'message': '"Mozilla/5.0 (X11; Linux x86_64; rv:38.0) Gecko/20100101 Firefox/38.0"',
        'agent': '"Mozilla/5.0 (X11; Linux x86_64; rv:38.0) Gecko/20100101 Firefox/38.0"',
    }

After the parser runs, the event will become:

.. code-block:: python

    {
        'hostname': 'server_hostname',
        'time_stamp': '2015-01-01T01:00:00',
        'message': '"Mozilla/5.0 (X11; Linux x86_64; rv:38.0) Gecko/20100101 Firefox/38.0"',
        'agent': '"Mozilla/5.0 (X11; Linux x86_64; rv:38.0) Gecko/20100101 Firefox/38.0"',
        'user_agent': {
            'os': {
                'family': 'Linux',
                'version': ''
            },
            'browser': {
                'family': 'Firefox',
                'version': '38'
            },
            'device': {
                'brand': None,
                'family': 'Other',
                'model': None,
            },
            'is_bot': False,
            'is_mobile': False,
            'is_pc': True,
            'is_tablet': False,
            'is_touch_capable': False,
        },
    }
'''
# pylint: enable=line-too-long
import user_agents
from . import ExtractFieldParser


class UserAgentParser(ExtractFieldParser):
    '''
    Extract OS and Broswer info from user agent string
    '''
    def __init__(self,
                 field='agent',
                 out_field=None,
                 *args,
                 **kwargs):
        super(UserAgentParser, self).__init__(field, *args, **kwargs)
        self.out_field = out_field

    def parse(self, event):
        ua_str = self.data
        ua = user_agents.parse(ua_str)
        ua_data = {
            'browser': {
                'family': ua.browser.family,
                'version': ua.browser.version_string,
            },
            'os': {
                'family': ua.os.family,
                'version': ua.os.version_string,
            },
            'device': {
                'family': ua.device.family,
                'brand': ua.device.brand,
                'model': ua.device.model,
            },
            'is_mobile': ua.is_mobile,
            'is_tablet': ua.is_tablet,
            'is_touch_capable': ua.is_touch_capable,
            'is_pc': ua.is_pc,
            'is_bot': ua.is_bot,
        }
        if self.out_field:
            event[self.out_field] = ua_data
        else:
            event.update(ua_data)
        return event


Parser = UserAgentParser



# vim: set et ts=4 sw=4 fileencoding=utf-8:
'''
yalp.test.testcase
==================
'''

import sys
import unittest
from ..config import settings
from ..pipeline.tasks import app

from .utils import override_settings


class YalpTestCase(unittest.TestCase):
    '''
    Base Yalp Test case. Handles settings up settings and allowing for
    easily overriding.
    '''
    _overridden_settings = None

    @classmethod
    def setUpClass(cls):
        if cls._overridden_settings:
            cls._cls_overridden_context = override_settings(
                **cls._overridden_settings)  # pylint: disable=not-a-mapping
            cls._cls_overridden_context.enable()

    @classmethod
    def tearDownClass(cls):
        if hasattr(cls, '_cls_overridden_context'):
            cls._cls_overridden_context.disable()
            delattr(cls, '_cls_overridden_context')

    def __call__(self, result=None, **kwargs):
        '''
        Wrapper around default __call__ method to perform common Yalp test
        set up. This means that user-defined Test Cases aren't required to
        include a call to super().setUp().
        '''
        testMethod = getattr(             # pylint: disable=C0103
            self, self._testMethodName)
        skipped = (getattr(self.__class__, "__unittest_skip__", False) or
                   getattr(testMethod, "__unittest_skip__", False))

        if not skipped:
            try:
                self._pre_setup()
            except Exception:  # pylint: disable=W0703
                result.addError(self, sys.exc_info())
                return
        super(YalpTestCase, self).__call__(result)
        if not skipped:
            try:
                self._post_teardown()
            except Exception:  # pylint: disable=W0703
                result.addError(self, sys.exc_info())
                return

    def _pre_setup(self):
        '''
        Performs any pre-test setup. Setting CELERY_ALWAYS_EAGER and
        getting the celery app with the current settings.
        '''
        settings.celery_advanced = {
            'CELERY_ALWAYS_EAGER': True,
        }
        self.app = app

    def _post_teardown(self):
        '''
        Performs any post-test things.
        '''
        pass



# vim: set et ts=4 sw=4 fileencoding=utf-8:
'''
yalp.outputs
============
'''
import logging
from ..pipeline import CeleryPipeline


class BaseOutputer(CeleryPipeline):
    '''
    Base outputer.
    '''
    def __init__(self, *args, **kwargs):
        super(BaseOutputer, self).__init__(*args, **kwargs)
        self.logger = logging.getLogger(__name__)

    def process_event(self, event):
        return self.output(event)

    def output(self, event):
        '''
        Output the event.
        '''
        raise NotImplementedError

    def shutdown(self):
        '''
        Shutdown the outputer, closing resources.
        '''
        raise NotImplementedError



# vim: set et ts=4 sw=4 fileencoding=utf-8:
'''
yalp.outputs.elasticsearch
==========================

The elasticsearch outputer sends events to an elasticsearch index.

.. warning:: This requires the pyelasticsearch_ pacakge to be installed.

This outputer supports the following configuration items:

*uri*
    The elasticsearch connection uri Formatted as
    ``http[s]://[user:password@]<host>[:port]/[path]``. Can also be a
    list of connection uris. Defaults to ``http://localhost:9200/``.

*index*
    The index name to store the documents. Default to
    ``yalp-%Y.%m.%d``.  The index can contain a `date format`_ string
    for a dynamic index.

*doc_type*
    The document name. Default to ``logs``.

*time_based*
    If the index is time based. This requires that the index name
    contains a date format string and that the event contains a valid
    time stamp. Default to ``True``.

*time_stamp_fmt*
    The date format of the time stamp in the event. Not used if the
    ``time_stamp`` field is a datetime. Default to ``%Y-%m-%dT%H:%M:%S``.

*manage_template*
    Allow yalp to manage the elasticsearch index template. Default to
    ``True``.

*template_name*
    The name of the index template to create. Default to ``yalp``.

*template_overwrite*
    Allow yalp to write over any existing template. Default to
    ``False``.

*buffer_size*
    The outputer will buffer this many events before sending them all to
    elasticsearch via a bulk insert. Default is ``500``.

*type*
    A type filter. Only output events of this type.

Example configuration.

.. code-block:: yaml

    outputs:
      - elasticsearch:
          uri: 'http://localhost:9200/'
          index: "yalp-%Y.%m.%d"
          doc_type: logs

.. _pyelasticsearch: https://pypi.python.org/pypi/pyelasticsearch/
.. _date format: https://docs.python.org/2/library/datetime.html#strftime-and-strptime-behavior
'''
from __future__ import absolute_import
from datetime import datetime

try:
    from elasticsearch import Elasticsearch
    from elasticsearch.exceptions import TransportError
    from elasticsearch.helpers import bulk, BulkIndexError
except ImportError:  # pragma: no cover
    pass
from ..exceptions import OutputException
from . import BaseOutputer


TEMPLATE = {
    'template': 'yalp-*',
    'settings': {
        'index.refresh_interval': '5s',
    },
    'mappings': {
        '_default_': {
            '_all': {
                'enabled': True,
                'omit_norms': True,
            },
            'dynamic_templates': [
                {
                    'message_field': {
                        'match': 'message',
                        'match_mapping_type': 'string',
                        'mapping': {
                            'type': 'string',
                            'index': 'analyzed',
                            'omit_norms': True,
                        },
                    },
                },
                {
                    'string_fields': {
                        'match': '*',
                        'match_mapping_type': 'string',
                        'mapping': {
                            'type': 'string',
                            'index': 'analyzed',
                            'omit_norms': True,
                            'fields': {
                                'raw': {
                                    'type': 'string',
                                    'index':
                                    'not_analyzed',
                                    'ignore_above': 256,
                                    'doc_values': True,
                                },
                            },
                        },
                    },
                }
            ],
            'properties': {
                'geoip': {
                    'type': 'object',
                    'dynamic': True,
                    'properties': {
                        'location': {'type': 'geo_point'},
                    },
                }
            },
        },
    },
}


class ElasticSearchOutputer(BaseOutputer):
    '''
    Send output to elasticsearch
    '''
    def __init__(self,
                 uri='http://localhost:9200/',
                 index='yalp-%Y.%m.%d',
                 doc_type='logs',
                 manage_template=True,
                 template_name='yalp',
                 template_overwrite=False,
                 time_based=True,
                 time_stamp_fmt='%Y-%m-%dT%H:%M:%S',
                 buffer_size=500,
                 *args,
                 **kwargs):
        super(ElasticSearchOutputer, self).__init__(*args, **kwargs)
        sniff_settings = True
        if not isinstance(uri, list):
            sniff_settings = False
            uri = [uri]
        self.es = Elasticsearch(
            uri,
            sniff_on_start=sniff_settings,
            sniff_on_connection_fail=sniff_settings,
        )
        self.index = index
        self.doc_type = doc_type
        self.time_based = time_based
        self.time_stamp_fmt = time_stamp_fmt
        if manage_template:
            self.es.indices.put_template(
                name=template_name,
                body=TEMPLATE,
                create=not template_overwrite,
                ignore=400,
            )
        if not self.time_based:
            self.es.indices.create(index=self.index, ignore=400)
        self.buffer_size = buffer_size
        self.buffer = []

    def get_index(self, event):
        '''
        Get the correct index name for the event.
        '''
        if self.time_based:
            try:
                time_stamp = event['time_stamp']
                if isinstance(time_stamp, str):
                    time_stamp = datetime.strptime(
                        time_stamp,
                        self.time_stamp_fmt,
                    )
                return time_stamp.strftime(self.index)
            except KeyError:
                self.logger.error('Time based event without time stamp')
                raise OutputException
            except ValueError:
                self.logger.error('Time stamp invalid datetime format')
                raise OutputException
            except AttributeError:
                self.logger.error('Time stamp not datetime or str')
                raise OutputException
        else:
            return self.index

    def _flush_buffer(self):
        ''' Write out events to elasticsearch '''
        try:
            try:
                bulk(self.es, self.buffer)
            except BulkIndexError as exc:
                self.logger.error(
                    'Failed to insert some events: %s',
                    exc.errors)
            self.buffer = []
        except TransportError:
            self.logger.warn('Error connection to elasticsearch. Will retry.')

    def output(self, event):
        try:
            self.buffer.append({
                '_type': self.doc_type,
                '_index': self.get_index(event),
                '_source': event
            })
        except OutputException:
            self.logger.error('Output exception', exc_info=True)
        if len(self.buffer) >= self.buffer_size:
            self._flush_buffer()

    def shutdown(self):
        self._flush_buffer()
        self.es.indices.flush(
            index='_all',
            ignore_unavailable=True,
        )


Outputer = ElasticSearchOutputer



# vim: set et ts=4 sw=4 fileencoding=utf-8:
'''
yalp.parsers.passthrough
========================
'''
from . import BaseParser


class PassThroughParser(BaseParser):
    '''
    Print input
    '''
    def __init__(self, *args, **kwargs):
        super(PassThroughParser, self).__init__(*args, **kwargs)

    def parse(self, event):
        return event


Parser = PassThroughParser



# vim: set et ts=4 sw=4 fileencoding=utf-8:
'''
yalp.parsers.geoip
==================

Extract Geo location data from an IP address.

The parser supports the following configuration items:

.. warning::
    This parser requires the pygeoip_ package. The pygeoip package
    uses MaxMind's GeoIP dat files to get geo info from IP addresses.
    See http://dev.maxmind.com/geoip/legacy/geolite/ for more info.

.. note::
    The geohash_ package is nessecary for converting latitude/longitude
    into geohashes. If not installed, the parser will store the raw
    latitude and longitude.

**geoip_dat**
    Path to the MaxMind GeoIP City dat file.

*field*
    The field containing the IP address to parse. If the field is not
    found in the event, the event will be skipped. Defaults to
    ``clientip``.

*out_field*
    The field to set the Geo data to. Defaults to ``geoip``.

*use_hash*
    Store location as a geohash. Default is ``True``. If set to ``False``
    location will be stored as ['lat', 'lon'] pair. Ignored if `geohash`_
    is not installed.

*type*
    A type filter. Events not of this type will be skipped.


Example configuration.

.. code-block:: yaml

    parsers:
      - geoip:
          field: 'clientip'
          geoip_dat: '/usr/share/GeoLiteCity.dat'


.. _pygeoip: https://pypi.python.org/pypi/pygeoip/
.. _geohash: https://pypi.python.org/pypi/python-geohash
'''
try:
    import geohash
    HAS_GEOHASH = True
except ImportError:
    HAS_GEOHASH = False

from pygeoip import GeoIP, GeoIPError
from . import ExtractFieldParser


def _loc_geohash(geo_info):
    ''' Create location with geohash '''
    return geohash.encode(
        geo_info.pop('latitude'),
        geo_info.pop('longitude'),
    )


def _loc_point(geo_info):
    ''' Create location as point [lon, lat] '''
    return [
        geo_info.pop('longitude'),
        geo_info.pop('latitude'),
    ]


class GeoIPParser(ExtractFieldParser):
    '''
    Get geo info from IP address.
    '''
    def __init__(self,
                 field='clientip',
                 out_field='geoip',
                 geoip_dat='',
                 use_hash=True,
                 *args,
                 **kwargs):
        super(GeoIPParser, self).__init__(field, *args, **kwargs)
        self.out_field = out_field
        self.get_loc = _loc_geohash if HAS_GEOHASH and use_hash else _loc_point
        try:
            self.geoip = GeoIP(geoip_dat)
        except (IOError, GeoIPError) as exc:
            self.logger.error('Invalid GeoIP Database file: %s', geoip_dat)
            raise exc

    def parse(self, event):
        ip_addr = self.data
        try:
            geo_info = self.geoip.record_by_addr(ip_addr)
            if 'latitude' in geo_info and 'longitude' in geo_info:
                geo_info['location'] = self.get_loc(geo_info)
            event[self.out_field] = geo_info
        except (IndexError, TypeError):
            self.logger.warn('Failed to get Geo info from ip: %s', ip_addr)
        return event


Parser = GeoIPParser



# vim: set et ts=4 sw=4 fileencoding=utf-8:
'''
yalp.outputs.mongo
==================

The mongo outputer sends events to a mongo collection. Each event
is recorded as a new document in the collection.

.. warning:: This requires the pymongo_ pacakge to be installed.

This outputer supports the following configuration items:

**uri**
    The mongodb connection uri. Formatted as
    ``mongodb://[user:password@]<host>[:port]/[auth_database]``

**database**
    The database name to store the documents.

**collection**
    The collection name.

*type*
    A type filter. Only output events of this type.

Example configuration.

.. code-block:: yaml

    outputs:
      - mongo:
          uri: 'mongodb://localhost:27017/yalp'
          database: yalp
          collection: logs

.. _pymongo: https://pypi.python.org/pypi/pymongo/
'''
try:
    import pymongo
except ImportError:  # pragma: no cover
    pass
from . import BaseOutputer


class MongoDBOutputer(BaseOutputer):
    '''
    Send output to mongo.
    '''
    def __init__(self,
                 uri,
                 database,
                 collection,
                 *args,
                 **kwargs):
        super(MongoDBOutputer, self).__init__(*args, **kwargs)
        self.client = pymongo.MongoClient(uri)
        self.database = self.client[database]
        self.collection = self.database[collection]

    def output(self, event):
        self.collection.insert(event)

    def shutdown(self):
        self.client.fsync()
        self.client.close()


Outputer = MongoDBOutputer



# vim: set et ts=4 sw=4 fileencoding=utf-8:
'''
yalp.parsers.keyvalue
=====================

Extract key, value paired data.

The parser supports the following configuration items:

**field**
    The field containing the key value pairs to parse.

*out_field*
    Set this to a field to store the parsed pairs under. If not set, the
    new fields are added at the top level of the event.

*sep*
    The seperator between key and value in a pair. Defaults to ``:``

*pair_sep*
    The seperator between pairs of key/values. Defaults to a single space.

*type*
    A type filter. Events not of this type will be skipped.
'''
from . import ExtractFieldParser


class KeyValueParser(ExtractFieldParser):
    '''
    Extract key value data.
    '''
    def __init__(self,
                 field,
                 out_field=None,
                 sep=':',
                 pair_sep=' ',
                 *args,
                 **kwargs):
        super(KeyValueParser, self).__init__(field, *args, **kwargs)
        self.out_field = out_field
        self.sep = sep
        self.pair_sep = pair_sep

    def parse(self, event):
        pairs = self.data.split(self.pair_sep)  # pylint: disable=no-member
        keyvalues = {}
        for pair in pairs:
            key, value = pair.split(self.sep, 1)
            keyvalues[key] = value
        if self.out_field:
            event[self.out_field] = keyvalues
        else:
            event.update(keyvalues)
        return event


Parser = KeyValueParser



# vim: set et ts=4 sw=4 fileencoding=utf-8:
'''
yalp.parsers.grok
=================

Use `grok`_ to parse the event. Any matched fields from the grok pattern
will be added to the event.

This parser supports the following connfiguration items:

**pattern**
    A grok pattern to match. See `available patterns`_ for details.

*field*
    The field from the event to parse. Defaults to ``message``.

*type*
    A type filter. Events not of this type will be skipped.


Example configuration.

.. code-block:: yaml

    parsers:
      - grok:
        pattern: '%{IP:ip_addr} %{WORD:request_type} %{URIPATHPARAM:path}'

With an input event like the following:

.. code-block:: python

    {
        'message': '192.168.0.1 GET /index.html',
        'time_stamp': '2015-01-01T01:00:00',
        'hostname': 'server_hostname',
    }

After the parser runs, the event will become:

.. code-block:: python

    {
        'message': '192.168.0.1 GET /index.html',
        'time_stamp': '2015-01-01T01:00:00',
        'hostname': 'server_hostname',
        'ip_addr': '192.168.0.1',
        'request_type': 'GET',
        'path': '/index.html',
    }

.. _available patterns: https://github.com/garyelephant/pygrok/tree/master/pygrok/patterns
.. _grok: https://www.elastic.co/guide/en/logstash/current/plugins-filters-grok.html#plugins-filters-grok
'''
import yalp_grok as grok
from . import ExtractFieldParser


class GrokParser(ExtractFieldParser):
    '''
    Process input with grok pattern match.
    '''
    def __init__(self, field='message', pattern=None, *args, **kwargs):
        super(GrokParser, self).__init__(field, *args, **kwargs)
        self.pattern = pattern
        self.compiled_pattern = grok.compile_pattern(self.pattern)

    def parse(self, event):
        matches = grok.grok_search(self.data, self.compiled_pattern)
        if matches:
            event.update(matches)
        return event


Parser = GrokParser



# vim: set et ts=4 sw=4 fileencoding=utf-8:
'''
yalp.parsers.regex
==================

The regex parser applies a regex to the message of an event. Any named
components of the regex become new keys in the event dict with the
matched strings becoming the values.

.. note::
    The original message is removed from the event.

This parser supports the following configuration items:

**regex**
    The regex to apply.

*type*
    A type filter. Only apply the regex to events of this type.


Example configuration.

.. code-block:: yaml

    parsers:
      - regex:
          regex: '(?P<month>\\w+)\\s+(?P<day>\\d+)'

'''
import re
from . import ExtractFieldParser


class RegexParser(ExtractFieldParser):
    '''
    Process input with regex.
    '''
    def __init__(self, regex=None, *args, **kwargs):
        super(RegexParser, self).__init__(*args, **kwargs)
        self.regex = regex

    def parse(self, event):
        message = self.data
        match = re.match(self.regex, message)
        if match:
            event.update(match.groupdict())
        return event


Parser = RegexParser



# vim: set et ts=4 sw=4 fileencoding=utf-8:
'''
yalp.inputs.log_handler
=======================

The log handler inputer is for use within Python's `logging facility`_.
Therefore it uses a different configuration method than normal YALP
plugins. Instead of using the YALP `YAML` configuration file, this
inputer is configured in another project's python logging configuration.

For example, in another Python project:

.. code-block:: python

    LOGGING = {
        'version': 1,
        'disabled_existing_loggers': False,
        'handlers': {
            'yalp': {
                'level': 'INFO',
                'class': 'yalp.inputs.log_handler.YalpHandler',
                'type': 'my_package_logs',
                'pipeline': {
                    'broker_url': 'amqp://guest@guest@localhost:5672//',
                },
            },
        },
        'loggers': {
            'my_package': {
                'handlers': ['yalp']
                'level': 'INFO',
            },
        }
    }

The handler accepts the following optional fields:

*type*
    The type of the event for the parsers/outputers to filter on.

*pipeline*
    A dictionary of YALP configuration settings. The ``pipeline`` option
    accepts the following fields:

    *parsers*
        Boolean option. If true events will be sent to the parsers, otherwise
        they will be sent directly to the outputers. Default is ``False``.

    Additionaly, ``pipeline`` option for the handler accepts all of the YALP
    Celery :ref:`configuration` settings, such as the ``broker_url``, with
    the same defaults.


Then in the Python project, send a log message as follows:

.. code-block:: python

    logger.info('a log message', extra={'additional': 'data will be included'})

This will create an event like:

.. code-block:: python

    {
        'time_stamp': '2015-01-01T01:00:00',
        'message': 'a log message',
        'additional': 'data will be included',
        'logger': 'my_pacakge.my_module',
        'funcName': 'func_with_log',
        'levelname': 'INFO',
        'levelno': 20,
        'hostname': 'my_host',
    }

.. _logging facility: https://docs.python.org/3/library/logging.html
'''
from __future__ import absolute_import

import time
import logging

from ..config import settings
from ..utils import get_hostname

from . import InputerMixin, DEFAULT_DATE_FMT

RESEVERD = frozenset((
    'stack',
    'args',
    'name',
    'msg',
    'message',
    'exc_info',
    'created',
    'msecs',
    'relativeCreated',
    'tags',
))


class YalpHandler(logging.Handler, InputerMixin, object):
    '''
    Get input from python logging
    '''
    convert = time.gmtime

    def __init__(self,
                 pipeline=None,
                 *args,
                 **kwargs):  # pylint: disable=unused-argument
        pipeline = pipeline or {}
        settings.update(pipeline)
        self.type_ = kwargs.get('type')
        self.hostname = get_hostname()
        logging.Handler.__init__(
            self,
            level=kwargs.get('level', logging.NOTSET),
        )

    def _timestamp(self, created):
        ''' Convert created time to time_stamp '''
        converted_time = self.convert(created)
        return time.strftime(DEFAULT_DATE_FMT, converted_time)

    def emit(self, record):
        event = {}
        for key, val in vars(record).items():
            if key.startswith('_') or key in RESEVERD:
                continue
            event[key] = val

        event.update({
            'time_stamp': self._timestamp(record.created),
            'message': self.format(record),
            'logger': record.name,
        })

        self.enqueue_event(event)


Inputer = YalpHandler



# vim: set et ts=4 sw=4 fileencoding=utf-8:
'''
yalp.parsers.json
=================

Load json data.

The parser supports the following configuration items:

*field*
    The field containing the json string to load. Defaults to ``message``.

*out_field*
    Set this to a field to store the loaded dict. If not set, the new
    fields are added at the top level of the event.

*type*
    A type filter. Events not of this type will be skipped.
'''
from __future__ import absolute_import
import json
from . import ExtractFieldParser


class JsonParser(ExtractFieldParser):
    '''
    Load json data.
    '''
    def __init__(self,
                 field='message',
                 out_field=None,
                 *args,
                 **kwargs):
        super(JsonParser, self).__init__(field, *args, **kwargs)
        self.out_field = out_field

    def parse(self, event):
        json_str = self.data
        try:
            data = json.loads(json_str)
            if self.out_field:
                event[self.out_field] = data
            else:
                event.update(data)
        except ValueError:
            self.logger.error('Failed to load json')
        return event


Parser = JsonParser



# vim: set et ts=4 sw=4 fileencoding=utf-8:
'''
yalp.outputs.stdout
===================
'''
from __future__ import print_function

import sys
import json
from . import BaseOutputer


class StdOutOutputer(BaseOutputer):
    '''
    Print output
    '''
    def __init__(self, out=sys.stdout, *args, **kwargs):
        super(StdOutOutputer, self).__init__(*args, **kwargs)
        self.out = out

    def output(self, event):
        self.out.write(json.dumps(event))

    def shutdown(self):
        pass


Outputer = StdOutOutputer



# vim: set et ts=4 sw=4 fileencoding=utf-8:
'''
yalp.parsers.transform
======================

Used to convert a field in the event to a different built-in type.

The parser supports the following configuration items:

..note::
    If the field fails to be transformed, The parser will log an error
    and leave the field as it was originally.

**field**
    The field to convert. If the field is not found in the event, the
    event will be skipped.

**to**
    Convert the field into this type. Supported types are::

        int
        float
        str

*type*
    A type filter. Events not of this type will be skipped.


Example configuration.

.. code-block:: yaml

    parsers:
      - transform:
          field: response_time
          to: int
'''
from ..utils import nested_put
from . import ExtractFieldParser


def _to_int(field):
    ''' Transform to int '''
    return int(field)


def _to_float(field):
    ''' Transform to float '''
    return float(field)


def _to_str(field):
    ''' Transform to str '''
    return str(field)


TRANSFORM_MAP = {
    'int': _to_int,
    'float': _to_float,
    'str': _to_str,
}


class TransformParser(ExtractFieldParser):
    '''
    Transform fields into new field types.
    '''
    def __init__(self, field, to, *args, **kwargs):
        super(TransformParser, self).__init__(field, *args, **kwargs)
        self.to = to
        self.to_func = TRANSFORM_MAP[self.to]

    def parse(self, event):
        try:
            transformed = self.to_func(self.data)
        except ValueError:
            transformed = self.data
            self.logger.error('Failed to transform field %s to %s',
                              self.field,
                              self.to)
        nested_put(event, self.field, transformed)
        return event


Parser = TransformParser



# vim: set et ts=4 sw=4 fileencoding=utf-8:
'''
yalp.parsers.timestamp
======================

Used to set the event time_stamp from another field in the event.

The parser supports the following configuration items:

**field**
    The field to parse for a datetime. If the field is not found in the
    event, the event will be skipped.

*out_field*
    The field to write the parsed time stamp to. Defaults to
    ``time_stamp``.

*timestamp_fmt*
    The `date format`_ string to format the time stamp. Defaults to
    ``%Y-%m-%dT$H:%M:%S``.

*to_utc*
    Convert the timestamp to UTC after parsing. Defaults to ``True``.

*type*
    A type filter. Events not of this type will be skipped.


Examaple configuration.

.. code-block:: yaml

    parsers:
      - timestamp:
          field: date_field

.. _date format: https://docs.python.org/2/library/datetime.html#strftime-and-strptime-behavior
'''
from dateutil import tz
from dateutil.parser import parse as dt_parse

from . import ExtractFieldParser

DEFAULT_DATE_FMT = '%Y-%m-%dT%H:%M:%S'
UTC = tz.tzutc()


class TimeStampParser(ExtractFieldParser):
    '''
    Search for datetime string and set time_stamp.
    '''
    def __init__(self,
                 field='timestamp',
                 out_field='time_stamp',
                 timestamp_fmt=DEFAULT_DATE_FMT,
                 to_utc=True,
                 *args,
                 **kwargs):
        super(TimeStampParser, self).__init__(field, *args, **kwargs)
        self.out_field = out_field
        self.timestamp_fmt = timestamp_fmt
        self.to_utc = to_utc

    def parse(self, event):
        parsed_dt = dt_parse(self.data, fuzzy=True)
        # pylint: disable=no-member
        if self.to_utc:
            try:
                parsed_dt = parsed_dt.astimezone(UTC)
            except ValueError:
                self.logger.info('Failed to convert to UTC')
        event[self.out_field] = parsed_dt.strftime(self.timestamp_fmt)
        # pylint: enable=no-member
        return event


Parser = TimeStampParser



# vim: set et ts=4 sw=4 fileencoding=utf-8:
'''
yalp.inputs.file
================

The file inputer creates an event for each line in the file. It
continues to follow the file much like the unix ``tail -F`` command. It
will correctly follow the file event if the inode changes, like from
tools like logroate. Additionally, the inputer saves its last position
in the file so it does not reprocess lines if the service is
stopped/restarted.


This inputer supports the following configuration items:

**path**
    The path of the file to collect events from.

*type*
    The type of event for parsers/outputers to filter on.


Example configuration.

.. code-block:: yaml

    inputs:
      - file:
          path: /var/log/messages

'''
import os
import hashlib
import time
from ..config import settings
from . import BaseInputer


class FileInputer(BaseInputer):
    '''
    Get input from a file.
    '''
    line_terminators = ('\r\n', '\n', '\r')

    def __init__(self, path, *args, **kwargs):
        super(FileInputer, self).__init__(*args, **kwargs)
        self.path = path
        self.sincedb_dir = settings.home or os.environ['HOME']
        self.sincedb = os.path.join(
            self.sincedb_dir,
            '.sincedb_{0}'.format(
                hashlib.md5(self.path.encode('utf-8')).hexdigest())
        )

    def _setup(self):
        '''
        Setup file handle seeking to last known position.
        '''
        self.file = open(self.path, 'r')
        self.cur_inode = os.fstat(self.file.fileno()).st_ino
        if os.path.exists(self.sincedb):
            with open(self.sincedb, 'r') as sincedb_file:
                for line in sincedb_file.readlines():
                    inode, position = line.strip().split()
                    if int(inode) == self.cur_inode:
                        self.file.seek(int(position))

    def _cleanup(self):
        '''
        Cleanup file handle, record current position.
        '''
        self._write_sincedb()
        self.file.close()

    def _write_sincedb(self):
        ''' Write current state to sincedb '''
        with open(self.sincedb, 'w') as sincedb_file:
            sincedb_file.write('{0} {1}'.format(self.cur_inode,
                                                self.file.tell()))

    def _follow(self, delay=1.0):
        '''
        Iterator generator that returns lines as data is added to the
        file. Iteration will stop when thread is triggered to stop.

        Based on:
        http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/157035
        '''
        trailing = True

        while 1:
            if self.stopped:
                break
            where = self.file.tell()
            line = self.file.readline()
            if line:
                if trailing and line in self.line_terminators:
                    trailing = False
                    continue

                if line[-1] in self.line_terminators:
                    line = line[:-1]
                    if line[-1:] == '\r\n' and '\r\n' in self.line_terminators:
                        # found crlf
                        line = line[:-1]

                trailing = False
                yield line
            else:
                trailing = True
                if os.stat(self.path).st_ino != self.cur_inode:
                    new_file = open(self.path, 'r')
                    self.file.close()
                    self.file = new_file
                    self.cur_inode = os.fstat(self.file.fileno()).st_ino
                    self._write_sincedb()
                else:
                    self.file.seek(where)
                time.sleep(delay)

    def run(self):
        self._setup()
        for line in self._follow():
            event = {'message': line}
            self.enqueue_event(event)
        self._cleanup()


Inputer = FileInputer



# vim: set et ts=4 sw=4 fileencoding=utf-8:
'''
yalp.pipeline.tasks
===================
'''
from __future__ import absolute_import
from celery import Celery, Task, bootsteps, uuid
from kombu import Consumer, Exchange, Queue

from ..config import settings
from ..utils import get_yalp_class


class YalpOutputersConsumer(bootsteps.ConsumerStep):
    '''
    Yalp outputer consumer
    '''
    _config = None
    _outputers = None

    @property
    def config(self):
        '''
        Get the configuration.
        '''
        if self._config is None:
            self._config = settings
        return self._config

    @property
    def outputers(self):
        '''
        Get the list of output classes.
        '''
        if self._outputers is None:
            outputs = []
            for conf in self.config.outputs:
                for plugin, config in conf.items():
                    outputs.append(get_yalp_class(plugin, config, 'output'))
            self._outputers = outputs
        return self._outputers

    def get_consumers(self, channel):
        return [
            Consumer(
                channel,
                queues=[Queue(settings.output_queue,
                              Exchange(settings.output_queue),
                              settings.output_queue)],
                callbacks=[self.handle_message],
                accept=['pickle', 'json', 'msgpack', 'yaml'],
            )
        ]

    def handle_message(self, body, message):
        '''
        Process the message.
        '''
        for outputer in self.outputers:
            outputer.run(body['message'])
        message.ack()

    def shutdown(self, c):
        for outputer in self.outputers:
            outputer.shutdown()
        super(YalpOutputersConsumer, self).shutdown(c)


def lazy_update_app_config():
    '''
    Load settings into celery as late as possible.
    '''
    config_updates = {
        'BROKER_URL': settings.broker_url,
        'CELERY_ROUTES': {
            'yalp.pipeline.tasks.process_message': {
                'queue': settings.parser_queue,
            },
        },
    }
    config_updates.update(settings.celery_advanced)
    return config_updates


app = Celery()
app.add_defaults(lazy_update_app_config)


class PipelineTask(Task):
    '''
    Pipeline celery task.
    '''
    abstract = True
    _config = None
    _parsers = None
    _outputers = None
    _output_queue = None
    _output_exchange = None

    @property
    def config(self):
        '''
        Get the parsers configuration.
        '''
        if self._config is None:
            self._config = settings
        return self._config

    @property
    def parsers(self):
        '''
        Get the list of parser classes.
        '''
        if self._parsers is None:
            parsers = []
            for conf in self.config.parsers:
                for plugin, config in conf.items():
                    parsers.append(get_yalp_class(plugin, config, 'parser'))
            self._parsers = parsers
        return self._parsers

    def reload_config(self):
        '''
        Re-read config from settings
        '''
        self._config = None
        self._parsers = None


def process_output(event):
    '''
    Send the event to the output consumer.
    '''
    exchange = Exchange(settings.output_queue)
    queue = Queue(settings.output_queue,
                  exchange,
                  settings.output_queue)
    with app.producer_or_acquire(None) as producer:
        producer.publish(
            {
                'task': 'yalp.pipeline.process_output',
                'id': uuid(),
                'message': event
            },
            serializer=settings.celery_serializer,
            exchange=exchange,
            routing_key=settings.output_queue,
            declare=[queue],
            retry=True,
        )


@app.task(base=PipelineTask)
def process_message(event):
    '''
    Process a message using settings from config.

    event
        The event to process. A dict containing the message. For example:

        .. code-block:: python

            {
                'message': 'the message to parse',
                'time_stamp': '2015-01-01T01:00:00',
                'hostname': 'input_host',
            }
    '''
    for parser in process_message.parsers:
        event = parser.run(event)
    process_output(event)
    return event



# vim: set et ts=4 sw=4 fileencoding=utf-8:
'''
yalp.scripts
============
'''
from __future__ import print_function
from __future__ import absolute_import

import os
import sys
import time
import argparse

from . import version
from .config import settings
from .utils import get_yalp_class, get_hostname
from .pipeline.tasks import app, YalpOutputersConsumer
from .exceptions import ShutdownException


class BaseEntryPoint(object):
    '''
    Main Entry point.
    '''
    def __init__(self,
                 description=None,
                 argv=None,
                 *args,
                 **kwargs):  # pylint: disable=W0613
        self.argv = argv or sys.argv[:]
        self.prog_name = os.path.basename(self.argv[0])
        self.description = description or self.prog_name
        self.parser = argparse.ArgumentParser(description='YALP')

    def add_arguments(self):
        '''
        Add arguments to arg parser.
        '''
        self.parser.add_argument('-v', '--version',
                                 action='store_true',
                                 default=False,
                                 help='Display YALP version')
        self.parser.add_argument('-c', '--config',
                                 default=None,
                                 help='Specify alternative config')

    def execute(self):
        '''
        Execute command.
        '''
        self.add_arguments()
        self.options = self.parser.parse_args(self.argv[1:])
        if self.options.version:
            print(version.__version__)
            sys.exit(0)
        if self.options.config:
            os.environ['YALP_CONFIG_FILE'] = self.options.config
        self.app = app


class ParsersEntryPoint(BaseEntryPoint):
    '''
    Entry point for starting parser workers.
    '''
    def execute(self):
        super(ParsersEntryPoint, self).execute()
        self.app.worker_main([
            'yalp-parsers',
            '--concurrency={0}'.format(settings.parser_workers),
            '--queues={0}'.format(settings.parser_queue),
            '--hostname={0}-{1}'.format(
                get_hostname(),
                settings.parser_worker_name,
            )
        ])


class OutputersEntryPoint(BaseEntryPoint):
    '''
    Entry point for starting outputers workers.
    '''
    def execute(self):
        super(OutputersEntryPoint, self).execute()
        self.app.steps['consumer'].add(YalpOutputersConsumer)
        self.app.worker_main([
            'yalp-outputers',
            '--concurrency={0}'.format(settings.output_workers),
            '--hostname={0}-{1}'.format(
                get_hostname(),
                settings.output_worker_name,
            )
        ])


def sigterm_handler(signo, stack_frame):  # pylint: disable=W0613
    '''
    Catch signal and raise ShutdownException to preform cleanup.
    '''
    raise ShutdownException(signo)


class InputsEntryPoint(BaseEntryPoint):
    '''
    Entry point for starting inputers.
    '''
    _inputers = None

    def __init__(self, max_iterations=None, interval=1, *args, **kwargs):
        super(InputsEntryPoint, self).__init__(*args, **kwargs)
        self.max_iterations = max_iterations
        self.interval = interval

    @property
    def inputers(self):
        '''
        Get the list of inputer classes.
        '''
        if self._inputers is None:
            inputers = []
            for conf in settings.inputs:
                for plugin, config in conf.items():
                    inputers.append(get_yalp_class(plugin, config, 'input'))
            self._inputers = inputers
        return self._inputers

    def check_iterations(self):
        ''' Check if we should sleep again '''
        if self.max_iterations is not None:
            self.max_iterations -= 1
            if self.max_iterations <= 0:
                raise ShutdownException
        return True

    def execute(self):
        super(InputsEntryPoint, self).execute()
        import signal
        signal.signal(signal.SIGTERM, sigterm_handler)
        signal.signal(signal.SIGINT, sigterm_handler)
        try:
            for inputer in self.inputers:
                inputer.start()
            while self.check_iterations():
                time.sleep(self.interval)
        except ShutdownException:
            for inputer in self.inputers:
                inputer.stop()
            for inputer in self.inputers:
                inputer.join()



# vim: set et ts=4 sw=4 fileencoding=utf-8:
'''
yalp.outputs.file
=================

The file outputer writes events to a file. Each event is recorded as a
JSON string.

This outputer supports the following configuration items:

**path**
    The path of the file to write the events.

*type*
    A type filter. Only output events of this type.

Example configutation.

.. code-block:: yaml

    outputs:
      - file:
          path: /var/log/all_messages

'''
import json
from . import BaseOutputer


class FileOutputer(BaseOutputer):
    '''
    Write output to a file.
    '''
    def __init__(self, path=None, *args, **kwargs):
        super(FileOutputer, self).__init__(*args, **kwargs)
        self.path = path
        self.outfile = open(self.path, 'a')

    def output(self, event):
        self.outfile.write(json.dumps(event) + '\n')

    def shutdown(self):
        self.outfile.flush()
        self.outfile.close()


Outputer = FileOutputer



# vim: set et ts=4 sw=4 fileencoding=utf-8:
'''
yalp.test.utils
===============
'''
from functools import wraps
from ..config import settings, DEFAULT_OPTS


class UserSettingsHolder(object):
    '''
    Holder for user configured settings.
    '''
    def __init__(self, defaults=None):
        self.__dict__['_deleted'] = set()
        if not defaults:
            defaults = DEFAULT_OPTS
        self.default_settings = defaults

    def __getattr__(self, name):
        if name in self._deleted:
            raise AttributeError
        return getattr(self.default_settings, name)

    def __setattr__(self, name, value):
        self._deleted.discard(name)
        super(UserSettingsHolder, self).__setattr__(name, value)

    def __delattr__(self, name):
        self._deleted.add(name)
        if hasattr(self, name):
            super(UserSettingsHolder, self).__delattr__(name)

    def __dir__(self):
        return list(self.__dict__) + dir(self.default_settings)


# pylint: disable=W0212
class override_settings(object):  # pylint: disable=C0103
    '''
    Acts as either a decorator, or a context manager. If it's a decorator it
    takes a function and returns a wrapped function. If it's a contextmanager
    it's used with the ``with`` statement. In either event entering/exiting
    are called before and after, respectively, the function/block is executed.
    '''
    def __init__(self, **kwargs):
        self.options = kwargs

    def __enter__(self):
        self.enable()

    def __exit__(self, exc_type, exc_value, traceback):
        self.disable()

    def __call__(self, test_func):
        from .testcase import YalpTestCase
        if isinstance(test_func, type):
            if not issubclass(test_func, YalpTestCase):
                raise Exception(
                    'Only subclasses of YalpTestCase can be decorated '
                    'with override_settings')
            self.save_options(test_func)
            return test_func
        else:
            @wraps(test_func)
            def inner(*args, **kwargs):  # pylint: disable=C0111
                with self:
                    return test_func(*args, **kwargs)
        return inner

    def save_options(self, test_func):  # pylint: disable=C0111
        if test_func._overridden_settings is None:
            test_func._overridden_settings = self.options
        else:
            # Duplicate dict to prevent subclasses from altering their parent.
            test_func._overridden_settings = dict(
                test_func._overridden_settings, **self.options)

    def enable(self):
        ''' Enable overrides to settings.  '''
        override = UserSettingsHolder(settings._wrapped)
        for key, new_value in self.options.items():
            setattr(override, key, new_value)
        self.wrapped = settings._wrapped
        settings._wrapped = override

    def disable(self):
        ''' Disable overrides to settings. '''
        settings._wrapped = self.wrapped
        del self.wrapped
# pylint: enable=W0212



# vim: set et ts=4 sw=4 fileencoding=utf-8:
'''
yalp.version
============

Get version of yalp
'''
from __future__ import print_function

from distutils.version import StrictVersion  # pylint: disable=E0611,F0401
__version__ = str(StrictVersion('1.10'))
del StrictVersion

if __name__ == '__main__':  # pragma: no cover
    print(__version__)



# vim: set et ts=4 sw=4 fileencoding=utf-8:
'''
yalp.test
=========
'''
from yalp.test.testcase import YalpTestCase
from yalp.test.utils import override_settings


__all__ = [
    'YalpTestCase', 'override_settings'
]



# vim: set et ts=4 sw=4 fileencoding=utf-8:
'''
yalp.config
===========
'''
from __future__ import print_function

import os
from logging.config import dictConfig
import yaml


EMPTY = object()

DEFAULT_LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'simple': {
        },
    },
    'handlers': {
        'console': {
            'level': 'DEBUG',
            'class': 'logging.StreamHandler',
            'formatter': 'simple',
        },
    },
    'loggers': {
        'yalp.inputs': {
            'handlers': ['console'],
        },
        'yalp.parsers': {
            'handlers': ['console'],
        },
        'yalp.outputs': {
            'handlers': ['console'],
        },
    }
}

DEFAULT_OPTS = {
    'broker_url': 'amqp://guest:guest@localhost:5672//',
    'parser_queue': 'parsers',
    'output_queue': 'outputs',
    'parser_worker_name': 'parser-workers',
    'output_worker_name': 'output-workers',
    'parser_workers': 5,
    'output_workers': 1,
    'celery_advanced': {},
    'celery_serializer': 'pickle',
    'inputs': [],
    'parsers': [],
    'outputs': [],
    'home': None,
    'input_packages': ['yalp.inputs'],
    'parser_packages': ['yalp.parsers'],
    'output_packages': ['yalp.outputs'],
    'log_format': '%(name)s: %(levelname)s [%(module)s:%(lineno)s] %(message)s',
    'log_level': 'WARN',
    'logging': None,
}


def _read_conf_file(path):
    '''
    Parse yaml config file into dictionary.
    '''
    if path is None:
        return {}

    with open(path, 'r') as conf_file:
        try:
            conf_opts = yaml.safe_load(conf_file.read()) or {}
        except yaml.YAMLError as err:
            print('Error parsing configuration file: %s - %s' % (path, err))
            conf_opts = {}
        return conf_opts


def load_config(path, defaults=None):
    '''
    Read in config file.
    '''
    if defaults is None:
        defaults = DEFAULT_OPTS

    overrides = _read_conf_file(path)
    opts = defaults.copy()
    if overrides:
        opts.update(overrides)
    return opts


# pylint: disable=W0212
def new_method_proxy(func):
    ''' Proxy function call to lazy get attrs '''
    def inner(self, *args):  # pylint: disable=C0111
        if self._wrapped is EMPTY:
            self._setup()
        return func(self._wrapped, *args)
    return inner
# pylint: enable=W0212


class LazyObject(object):
    '''
    Wrapper for another class to delay instantiation.
    '''
    _wrapped = None

    def __init__(self):
        self._wrapped = EMPTY

    __getattr__ = new_method_proxy(getattr)

    def __setattr__(self, name, value):
        if name == '_wrapped':
            self.__dict__['_wrapped'] = value
        else:
            if self._wrapped is EMPTY:
                self._setup()
            setattr(self._wrapped, name, value)

    def __delattr__(self, name):
        if name == '_wrapped':
            raise TypeError('can\'t delete _wrapped.')
        if self._wrapped is EMPTY:
            self._setup()
        delattr(self._wrapped, name)

    def _setup(self):
        '''
        Must be implemented by subclasses to initialize wrapped object.
        '''
        raise NotImplementedError

    __dir__ = new_method_proxy(dir)


class LazySettings(LazyObject):
    '''
    Delay loading of settings.
    '''
    def _setup(self):
        settings_file = os.environ.get('YALP_CONFIG_FILE', None)
        self._wrapped = Settings(settings_file)
        self._configure_logging()

    def _configure_logging(self):
        ''' Configure logger from settings '''
        defauls = DEFAULT_LOGGING.copy()
        defauls['formatters']['simple']['format'] = self.log_format
        defauls['loggers']['yalp.inputs']['level'] = self.log_level
        defauls['loggers']['yalp.parsers']['level'] = self.log_level
        defauls['loggers']['yalp.outputs']['level'] = self.log_level
        dictConfig(defauls)
        if self.logging:
            dictConfig(self.logging)

    def update(self, update_dict):
        ''' Update settings from dict '''
        for key, val in update_dict.items():
            self.__setattr__(key, val)


class Settings(object):
    '''
    Load settings from yaml into object.
    '''
    def __init__(self, settings_file, defaults=None):
        opts = load_config(settings_file, defaults=defaults)
        for opt, value in opts.items():
            setattr(self, opt, value)


settings = LazySettings()



#!/usr/bin/env python
# vim: set et ts=4 sw=4 fileencoding=utf-8:
'''
Setup script for yalp
'''

import os
from setuptools import setup, find_packages

# Ensure we are in yalp source dir
SETUP_DIRNAME = os.path.dirname(__file__)
if SETUP_DIRNAME != '':
    os.chdir(SETUP_DIRNAME)

YALP_VERSION = os.path.join(os.path.abspath(SETUP_DIRNAME),
                            'yalp',
                            'version.py')
YALP_REQS = os.path.join(os.path.abspath(SETUP_DIRNAME),
                         'requirements.txt')

# pylint: disable=W0122
exec(compile(open(YALP_VERSION).read(), YALP_VERSION, 'exec'))
# pylint: enable=W0122

VER = __version__  # pylint: disable=E0602

REQUIREMENTS = []
with open(YALP_REQS) as rfh:
    for line in rfh.readlines():
        if not line or line.startswith('#'):
            continue
        REQUIREMENTS.append(line.strip())


SETUP_KWARGS = {
    'name': 'yalp',
    'version': VER,
    'url': 'https://github.com/yalp-log/yalp',
    'license': 'Apache-2',
    'description': 'Distributed log parsing and collection.',
    'author': 'Timothy Messier',
    'author_email': 'tim.messier@gmail.com',
    'classifiers': [
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Environment :: Console',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: POSIX :: Linux',
    ],
    'packages': find_packages(exclude=[
        '*.tests*', '*.tests.*', 'tests.*', 'tests',
    ]),
    'package_data': {},
    'data_files': [],
    'scripts': [
        'scripts/yalp',
        'scripts/yalp-inputs',
        'scripts/yalp-parsers',
        'scripts/yalp-outputers',
    ],
    'install_requires': REQUIREMENTS,
    'zip_safe': False,
}

if __name__ == '__main__':
    setup(**SETUP_KWARGS)
